#!/usr/bin/env python
import hysop
from hysop.constants import np, HDF5
from hysop.mpi.main_var import MPI, main_comm
from hysop.tools.parameters import Discretization, IOParams
from hysop.methods_keys import Scales, MultiScale
from hysop.operator.advection import Advection
from hysop.operator.analytic import Analytic
from hysop.operator.custom import CustomMonitor
from hysop.operator.hdf_io import HDF_Writer
from hysop.problem.problem import Problem
from hysop.problem.simulation import Simulation
sin, cos, pi, sqrt = np.sin, np.cos, np.pi, np.sqrt


def scalaire(res, x, y, z, t=0.):
    r = sqrt((x - 0.35) ** 2 + (y - 0.35) ** 2 + (z - 0.35) ** 2)
    res[0][...] = 0.
    res[0][r < 0.15] = 1.
    return res


def vitesse(res, x, y, z, t=0.):
    res[0][...] = 2. * sin(pi*x)**2*sin(2.*pi*y)*sin(2.*pi*z)*cos(t*pi/3.)
    res[1][...] = -sin(2.*pi*x)*sin(pi*y)**2*sin(2.*pi*z)*cos(t*pi/3.)
    res[2][...] = -sin(2.*pi*x)*sin(2.*pi*y)*sin(pi*z)**2*cos(t*pi/3.)
    return res


def volume(_simu, var):
    v_loc = np.sum(var[0].data[0] > 0.5) * np.prod(
        var[0].topology.mesh.space_step)
    return main_comm.allreduce(sendobj=v_loc, op=MPI.SUM)


# Parameters
dim = 3
boxLength = [1., ] * dim
boxMin = [0., ] * dim
nbElem_v = [65] * dim
nbElem_s = [65] * dim
dv = Discretization(nbElem_v)
ds = Discretization(nbElem_s)

timeStep = 0.35 / (4. * np.pi)
finalTime = 3.
outputModulo = 10
simu = Simulation(tinit=0.0, tend=finalTime,
                  timeStep=timeStep, iterMax=120)

# Domain
box = hysop.Box(length=boxLength, origin=boxMin)

# Fields
scal = hysop.Field(domain=box, name='Scalar', formula=scalaire)
velo = hysop.Field(domain=box, name='Velocity',
                   is_vector=True, formula=vitesse)
# Operators
advec = Advection(velo,
                  discretization=dv,
                  variables={scal: ds},
                  method={Scales: 'p_64', MultiScale: 'L4_4'})
advec.discretize()

# Get topologies from operator
topo_v = advec.discreteFields[velo].topology
topo_s = advec.discreteFields[scal].topology

velocity = Analytic(variables={velo: topo_v})

volume_m = CustomMonitor(function=volume, res_shape=1,
                         variables={scal: topo_s},
                         io_params=IOParams(filename="volume.dat"))
p = HDF_Writer(variables={scal: topo_s},
               io_params=IOParams(frequency=outputModulo,
                                  filename="levelset",
                                  fileformat=HDF5))
pb = Problem([velocity, advec, p, volume_m],
             simu, dumpFreq=-1)

pb.setup()
scal.initialize(topo=topo_s)
velo.initialize(topo=topo_v)

volume_m.apply(simu)
p.apply(simu)

ctime = MPI.Wtime()
pb.solve()
print MPI.Wtime() - ctime

simu.finalize()
p.apply(simu)
volume_m.apply(simu)

pb.finalize()
print pb.profiler
