"""
@file mesh.py
Cartesian mesh class for local mesh.

"""
from hysop.constants import debug
import hysop.tools.numpywrappers as npw
from hysop.tools.parameters import Discretization
import numpy as np


class SubMesh(object):
    """
    Description of a local mesh, on one process of
    the topology.
    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, domain, discretization, global_start, resolution):
        """
        Example of indices for a 8 point domain with 2 ghosts and 2 procs.
        \code
        X : ghost point
        o : compute point

           |o o o o o o o o|      domain
            0             7       global indices


        X X o o o o X X           -> Proc no1
                X X o o o o X X   -> Proc no2

        Proc no1:
        resolution : (8,)
        global start/end : 0/3
        local start/end : 2/5

        Proc no2:
        resolution: (8,)
        global start/end : 4/7
        local start/end : 2/5
        \endcode
        """
        ## Local resolution of this mesh, INCLUDING ghost points
        self.resolution = npw.asdimarray(resolution)
        ## Dimension of the mesh
        self.dim = self.resolution.size
        ## index of the lowest "computed" point of this mesh
        ## (in each dir) in the global mesh
        self.global_start = npw.asdimarray(global_start)
        assert isinstance(discretization, Discretization)
        self.discretization = discretization
        ghosts = self.discretization.ghosts
        ## index of the upper point (in each dir), global mesh
        self.global_end = self.global_start + self.resolution - 1 - 2 * ghosts
        ## Mesh step size
        global_resolution = self.discretization.resolution
        self.space_step = npw.asarray(domain.length / (global_resolution - 1))
        ## Mesh local indices, only for "computed" points
        ## (i.e. excluding ghosts)
        self.local_start = ghosts.copy()
        self.local_end = self.resolution - ghosts - 1
        ## List of indices for computational points (i.e. without ghosts)
        ## usage : field[iCompute] returns the array field for all indices
        ## excluding ghost points.
        self.iCompute = []
        ## List of indices to be used for integration over the domain.
        for d in range(self.dim):
            self.iCompute.append(slice(self.local_start[d],
                                       self.local_end[d] + 1))
        self.allgrid = []
        for d in range(self.dim):
            self.allgrid.append(slice(0, self.resolution[d]))

        ## Coordinates of the "lowest" point of this mesh (including ghost)
        ## Warning FP : this strongly depends on where is the origin
        ## of the domain, of the type of boundary conditions
        ## and if origin is on ghosts or "real" points.
        self.origin = domain.origin.copy()
        self.origin[:] += self.space_step[:] * (self.global_start[:]
                                                - ghosts[:])

        self.end = self.origin + self.space_step * (self.resolution - 1)
        if self.dim == 1:
            cx = np.linspace(self.origin, self.end, self.resolution)
            self.coords = tuple([cx, ])
        elif self.dim == 2:
            cx = np.linspace(self.origin[0], self.end[0],
                          self.resolution[0])[:, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                          self.resolution[1])[np.newaxis, :]
            self.coords = tuple([cx, cy])
        elif self.dim == 3:
            cx = np.linspace(self.origin[0], self.end[0],
                          self.resolution[0])[:, np.newaxis, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                          self.resolution[1])[np.newaxis, :, np.newaxis]
            cz = np.linspace(self.origin[2], self.end[2],
                          self.resolution[2])[np.newaxis, np.newaxis, :]
            self.coords = tuple([cx, cy, cz])

    def indices(self, tab):
        """
        returns indices of the point of coordinates (close to) tab = x, y, z
        If (x, y, z) is not a grid point, it returns the closest grid point.
        """
        ind = []
        for d in xrange(self.dim):
            cond = np.where(abs(self.coords[d] - tab[d])
                            < (self.space_step[d] * 0.5))
            if cond[0].size > 0:
                ind.append(cond[d][0])
            else:
                return False
        return tuple(ind)

    def __str__(self):
        """
        Sub mesh display
        """
        s = 'Ghost layer : ' + str(self.discretization.ghosts) + '\n'
        s += 'Global mesh resolution : ' + str(self.discretization.resolution)
        s += '\n'
        s = ' Sub mesh resolution:' + str(self.resolution) + '\n'
        s += 'Starting point global indices :' + str(self.global_start) + '\n'
        s += 'Local indices for "computed" points (excluding ghosts) :\n'
        s += '   start = ' + str(self.local_start)
        s += ', end = ' + str(self.local_end) + '\n'
        return s

    def toIndexLocal(self, glob_index):
        """
        Return the local mesh indices from
        their global indices.
        Ilocal = Iglobal - _global_start + ghost
        @param glob_index an array of size domain.dim*2
        with glob_index = [start_dir1, end_dir1, start_dir2, end_dir2, ...]
        return an array of indices, same 'setup' as glob_index.
        """
        ilocal = npw.asarray(glob_index).copy()
        ghosts = self.discretization.ghosts
        ilocal[0::2] = ilocal[0::2] - self.global_start + ghosts
        ilocal[1::2] = ilocal[1::2] - self.global_start + ghosts
        return ilocal

    def toIndexLocalFull(self, sl):
        """
        Return the local mesh indices from
        their global indices.
        Ilocal = Iglobal - _global_start + ghost
        @param glob_index an array of size domain.dim*2
        with glob_index = [start_dir1, end_dir1, start_dir2, end_dir2, ...]
        return an array of indices, same 'setup' as glob_index.
        """
        shift = self.local_start - self.global_start
        imax = self.local_end + self.discretization.ghosts
        dimension = self.discretization.ghosts.size
        return [slice(max(sl[i].start + shift[i], 0),
                      min(sl[i].stop + shift[i], imax[i]))
                for i in xrange(dimension)]

    def toIndexLocal2(self, sl):
        """
        Return the local mesh indices from
        their global indices.
        Ilocal = Iglobal - _global_start + ghost
        @param sl: the list of slices (one per dir) for global indices
        something like [(start_dir1, end_dir1), ...]
        @return the same kind of list but in local coordinates
        """
        shift = self.local_start - self.global_start
        imax = self.local_end + 1
        imin = self.local_start
        dimension = self.discretization.ghosts.size
        return [slice(max(sl[i].start + shift[i], imin[i]),
                      min(sl[i].stop + shift[i], imax[i]))
                for i in xrange(dimension)]

    def toIndexGlobal(self, sl):
        """
        Return the global mesh indices from
        their local indices.
        Iglobal = Ilocal + _global_start - ghost
        @param loc_index an array of size domain.dim*2
        with loc_index = [start_dir1, end_dir1, start_dir2, end_dir2, ...]
        return an array of indices, same 'setup' as loc_index.
        """
        shift = self.global_start - self.local_start
        dimension = self.discretization.ghosts.size
        return [slice(sl[i].start + shift[i], sl[i].stop + shift[i])
                for i in xrange(dimension)]

    def __eq__(self, other):
        """
        = operator for mesh.
        We consider that 2 submeshes are equal if all the conditions
        below are fulfilled:
        - their global resolution is the same
        - the ghost layer is the same
        - their local resolution is the same
        """
        if self.__class__ != other.__class__:
            return False
        return self.discretization == other.discretization and\
            npw.equal(self.resolution, other.resolution).all()
