# -*- coding: utf-8 -*-
import hysop as pp
import numpy as np
from hysop.fields.continuous import Field
import hysop.numerics.differential_operations as diffop
import hysop.tools.numpywrappers as npw
import math as m
pi = m.pi
cos = np.cos
sin = np.sin


## Function to compute TG velocity
def computeVel(res, x, y, z, t):
    res[0][...] = sin(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


## Function to compute reference vorticity
def computeVort(res, x, y, z, t):
    res[0][...] = - cos(x) * sin(y) * sin(z)
    res[1][...] = - sin(x) * cos(y) * sin(z)
    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
    return res


def computeVel2(res, x, y, t):
    res[0][...] = sin(x) * cos(y)
    res[1][...] = - cos(x) * sin(y)
    return res


def computeVort2(res, x, y, t):
    res[0][...] = 2. * sin(x) * sin(y)
    return res


def analyticalDivWV(res, x, y, z, t):
    res[0][...] = - sin(y) * cos(y) * sin(z) * cos(z) * \
        (- sin(x) * sin(x) + cos(x) * cos(x)) + \
        2. * cos(x) * cos(x) * sin(z) * cos(z) * sin(y) * cos(y)
    res[1][...] = - 2. * cos(y) * cos(y) * sin(z) * cos(z) * sin(x) * cos(x) +\
        sin(x) * cos(x) * sin(z) * cos(z) * (cos(y) * cos(y) - sin(y) * sin(y))
    res[2][...] = 0.
    return res


def analyticalGradVxW(res, x, y, z, t):
    res[0][...] = - sin(y) * cos(y) * sin(z) * cos(z)
    res[1][...] = sin(x) * cos(x) * sin(z) * cos(z)
    res[2][...] = 0.
    return res


def analyticalDivStressTensor(res, x, y, z, t):
    res[0][...] = - 3. * sin(x) * cos(y) * cos(z)
    res[1][...] = 3. * cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res

def analyticalDivAdvection(res, x, y, z, t):
    res[0][...] = - cos(z) * cos(z) * (cos(x) * cos(x) - sin(x) * sin(x)) - \
                  cos(z) * cos(z) * (cos(y) * cos(y) - sin(y) * sin(y))
    return res


nb = 65
from hysop.tools.parameters import Discretization
d3 = Discretization([nb] * 3, [2] * 3)
d2 = Discretization([nb] * 2, [2] * 2)


def init(discr, vform, wform):
    dim = len(discr.resolution)
    box = pp.Box(length=[2.0 * pi] * dim)
    topo = box.create_topology(discretization=discr)
    velo = Field(domain=box, formula=vform,
                 name='Velocity', is_vector=True)
    vorti = Field(domain=box, formula=wform,
                  name='Vorticity', is_vector=dim == 3)
    velo.discretize(topo)
    vorti.discretize(topo)
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)
    wd = vorti.discreteFields[topo]
    vd = velo.discreteFields[topo]
    return vd, wd


def testCurl():
    vd, wd = init(d3, computeVel, computeVort)
    memshape = vd.data[0].shape
    result = [npw.zeros(memshape) for i in xrange(3)]
    lwork = diffop.Curl.getWorkLengths()
    work = [npw.zeros(memshape) for i in xrange(lwork)]
    curlOp = diffop.Curl(vd.topology, work)
    result = curlOp(vd.data, result)
    ind = vd.topology.mesh.iCompute
    for i in xrange(3):
        assert np.allclose(wd.data[i][ind], result[i][ind])


def testCurl2D():
    vd, wd = init(d2, computeVel2, computeVort2)
    memshape = vd.data[0].shape
    result = [npw.zeros(memshape)]
    lwork = diffop.Curl.getWorkLengths()
    work = [npw.zeros(memshape) for _ in xrange(lwork)]
    curlOp = diffop.Curl(vd.topology, work)
    result = curlOp(vd.data, result)
    ind = vd.topology.mesh.iCompute
    assert np.allclose(wd.data[0][ind], result[0][ind])


def testDivWV():
    vd, wd = init(d3, computeVel, computeVort)
    box = vd.domain
    topo = vd.topology
    # Reference field
    ref = Field(domain=box, formula=analyticalDivWV,
                name='Analytical', is_vector=True)
    ref.discretize(topo)
    ref.initialize(topo=topo)
    refd = ref.discreteFields[topo]
    memshape = vd.data[0].shape
    # Div operator
    lwork = diffop.DivWV.getWorkLengths()
    work = [npw.zeros(memshape) for i in xrange(lwork)]
    divOp = diffop.DivWV(topo, work)
    result = [npw.zeros(memshape) for i in xrange(3)]
    result = divOp(vd.data, wd.data, result)

    # Numerical VS analytical
    Lx = box.length[0]
    errX = (Lx / (nb - 1)) ** 4
    ind = topo.mesh.iCompute
    for i in xrange(3):
        assert np.allclose(refd[i][ind], result[i][ind], rtol=errX)


def testGradVxW():
    vd, wd = init(d3, computeVel, computeVort)
    box = vd.domain
    topo = vd.topology
    # Reference field
    ref = Field(domain=box, formula=analyticalGradVxW,
                name='Analytical', is_vector=True)
    ref.discretize(topo)
    ref.initialize(topo=topo)
    refd = ref.discreteFields[topo]
    memshape = vd.data[0].shape
    lwork = diffop.GradVxW.getWorkLengths()
    work = [npw.zeros(memshape) for i in xrange(lwork)]
    gradOp = diffop.GradVxW(topo, work)
    result = [npw.zeros(memshape) for i in xrange(3)]
    diag = npw.zeros(2)
    result, diag = gradOp(vd.data, wd.data, result, diag)
    # Numerical VS analytical
    errX = (box.length[0] / (nb - 1)) ** 4
    ind = topo.mesh.iCompute
    for i in xrange(3):
        assert np.allclose(refd[i][ind], result[i][ind], rtol=errX)

# test for RHS of Pressure-Poisson equation
def testDivAdvection():
    dim = len(d3.resolution)
    box = pp.Box(length=[2.0 * pi] * dim)
    topo = box.create_topology(discretization=d3)
    # Velocity vector field
    velo = Field(domain=box, formula=computeVel,
                 name='Velocity', is_vector=True)
    velo.discretize(topo)
    velo.initialize(topo=topo)
    vd = velo.discreteFields[topo]
    # Reference scalar field
    ref = Field(domain=box, formula=analyticalDivAdvection,
                name='Analytical')
    ref.discretize(topo)
    ref.initialize(topo=topo)
    refd = ref.discreteFields[topo]
    memshape = vd.data[0].shape
    divAdvOp = diffop.DivAdvection(topo)
    result = [npw.zeros(memshape)]
    result = divAdvOp(vd.data, result)
    # Numerical VS analytical
    errX = (box.length[0] / (nb - 1)) ** 4
    ind = topo.mesh.iCompute
    assert np.allclose(refd[0][ind], result[0][ind], rtol=errX)


if __name__ == "__main__":
    testCurl()
    testCurl2D()
    testDivWV()
    testGradVxW()
    testDivAdvection()
