"""
@file operator/differential.py

Differential operators
"""
from hysop.constants import debug
from hysop.operator.computational import Computational
from hysop.operator.discrete.differential import CurlFFT, CurlFD,\
    GradFD, DivAdvectionFD
from hysop.methods_keys import SpaceDiscretisation
from hysop.operator.continuous import opsetup
from hysop.numerics.finite_differences import FD_C_4,\
    FD_C_2, FiniteDifference
import hysop.default_methods as default
from abc import ABCMeta, abstractmethod


class Differential(Computational):
    __metaclass__ = ABCMeta

    ## @debug
    ## def __new__(cls, *args, **kw):
    ##     return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, invar, outvar, **kwds):
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(Differential, self).__init__(variables=[invar, outvar], **kwds)
        if self.method is None:
            self.method = default.DIFFERENTIAL

        ## input variable
        self.invar = invar
        ## Curl of input
        self.outvar = outvar

        # Remark : at the time, all variables must have the same topology.
        # This is implicitely checked with the assert on kwds['variables']:
        # the only construction allowed is :
        # (invar= ..., outvar=..., discretization=...)
        self.output = [outvar]
        self.input = [invar]

    def discretize(self):

        if self.method[SpaceDiscretisation] is 'fftw':
            super(Differential, self)._fftw_discretize()

        elif self.method[SpaceDiscretisation] is FD_C_4:
            # Finite differences method
            # Minimal number of ghost points
            nbGhosts = 2
            super(Differential, self)._standard_discretize(nbGhosts)
        elif self.method[SpaceDiscretisation] is FD_C_2:
            nbGhosts = 1
            super(Differential, self)._standard_discretize(nbGhosts)

        else:
            raise ValueError("Unknown method for space discretization of the\
                differential operator.")

        msg = 'Operator not yet implemented for multiple resolutions.'
        assert self._single_topo, msg

    @abstractmethod
    @opsetup
    def setup(self, rwork=None, iwork=None):
        """
        Last step of initialization. After this, the operator must be
        ready for apply call.

        Main step : setup for discrete operators.
        """


class Curl(Differential):
    """
    Computes \f$ outVar = \nabla inVar \f$
    """
    def get_work_properties(self):
        if not self._is_discretized:
            msg = 'The operator must be discretized '
            msg += 'before any call to this function.'
            raise RuntimeError(msg)
        res = {'rwork': None, 'iwork': None}
        if self.method[SpaceDiscretisation].mro()[1] is FiniteDifference:
            from hysop.numerics.differential_operations \
                import Curl as NumCurl
            work_length = NumCurl.getWorkLengths()
            shape = self.discreteFields[self.invar].data[0].shape
            res['rwork'] = []
            for _ in xrange(work_length):
                res['rwork'].append(shape)
        return res

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        # Create a discrete operator, according to the chosen method
        if self.method[SpaceDiscretisation] is 'fftw':
            self.discrete_op = CurlFFT(
                invar=self.discreteFields[self.invar],
                outvar=self.discreteFields[self.outvar],
                method=self.method, rwork=rwork)
        elif self.method[SpaceDiscretisation].mro()[1] is FiniteDifference:
            self.discrete_op = CurlFD(
                invar=self.discreteFields[self.invar],
                outvar=self.discreteFields[self.outvar],
                method=self.method)
        else:
            raise ValueError("The required Space Discretisation is\
                not available for Curl.")
        self._is_uptodate = True


class Grad(Differential):
    """
    Computes \f$ outVar = \nabla inVar \f$
    """

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        if self.method[SpaceDiscretisation].mro()[1] is not FiniteDifference:
            raise ValueError("Grad operator only\
                implemented with finite differences. Please change\
                method[SpaceDiscretisation] value.")
        # Create a discrete operator, according to the chosen method
        # (only finite differences at the time).
        self.discrete_op = GradFD(invar=self.discreteFields[self.invar],
                                       outvar=self.discreteFields[self.outvar],
                                       method=self.method, rwork=rwork)
        self._is_uptodate = True


class DivAdvection(Differential):
    """
    Computes \f$ outVar = -\nabla .(invar . \nabla invar) \f$
    """

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        if self.method[SpaceDiscretisation].mro()[1] is not FiniteDifference:
            raise ValueError("DivAdvection operator only\
                implemented with finite differences. Please change\
                method[SpaceDiscretisation] value.")
        # Create a discrete operator, according to the chosen method
        # (only finite differences at the time).
        self.discrete_op = DivAdvectionFD(
            invar=self.discreteFields[self.invar],
            outvar=self.discreteFields[self.outvar],
            method=self.method, rwork=rwork)
        self._is_uptodate = True
