# -*- coding: utf-8 -*-
"""
@file operator/discrete/baroclinic.py
Discrete MultiPhase Rot Grad P
"""
from hysop.operator.discrete.discrete import DiscreteOperator
import hysop.numerics.differential_operations as diff_op
from hysop.constants import debug, XDIR, YDIR, ZDIR
from hysop.methods_keys import SpaceDiscretisation
from hysop.numerics.update_ghosts import UpdateGhosts
from hysop.tools.profiler import ftime
import hysop.tools.numpywrappers as npw


class Baroclinic(DiscreteOperator):
    """
    TODO : describe this operator ...
    """
    @debug
    def __init__(self, velocity, vorticity, density, viscosity,
                 formula=None, **kwds):
        """
        Constructor.
        Create the baroclinic term -GradRho/rho x GradP/rho
        in N.S equation
        @param velocity : discretization of the velocity field
        @param vorticity : discretization of the vorticity field
        @param density : discretization of a scalar field
        @param viscosity
        @param formula : formula to initialize u^(n-1)
        Note : this should be the formula used to initialize
        the velocity field
        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        if 'method' not in kwds:
            import hysop.default_methods as default
            kwds['method'] = default.BAROCLINIC

        super(Baroclinic, self).__init__(variables=[velocity, vorticity,
                                                    density], **kwds)
        self.velocity = velocity
        self.velocity_old = npw.zeros_like(self.velocity.data)
        self.vorticity = vorticity
        self.density = density
        self.viscosity = viscosity
        self.input = [self.velocity, self.vorticity, self.density]
        self.output = [self.vorticity]

        # prepare ghost points synchro for velocity (vector)
        # and density (scalar) fields
        self._synchronizeVel = UpdateGhosts(self.velocity.topology,
                                            self.velocity.nb_components)
        self._synchronizeRho = UpdateGhosts(self.density.topology,
                                            self.density.nb_components)

        # u^(n-1) initialization : u^(n-1) = u(t=0).
        time = 0.
        arg_list = self.velocity.topology.mesh.coords + (time,)
        self.velocity_old = formula(self.velocity_old, *arg_list)

    @debug
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        dt = simulation.timeStep
        # Synchronize ghost points of velocity and density
        self._synchronizeVel(self.velocity.data)
        self._synchronizeRho(self.density.data)

        topo = self.velocity.topology
        iCompute = topo.mesh.iCompute
        gradOp = diff_op.GradS(topo, self.method[SpaceDiscretisation])

        result = npw.zeros_like(self.velocity.data)

        # result = du/dt
        for d in xrange(self.velocity.dimension):
            result[d][iCompute] = (self.velocity[d][iCompute] -
                                   self.velocity_old[d][iCompute]) / dt

        # result = result + (u. grad)u
        # (u. grad)u = (u.du/dx + v.du/dy + w.du/dz ;
        #               u.dv/dx + v.dv/dy + w.dv/dz ;
        #               u.dw/dx + v.dw/dy + w.dw/dz)
        UgradU = npw.zeros_like(self.velocity.data)
        tempGrad = npw.zeros_like(self.velocity.data)
        tempGrad = gradOp(self.velocity[XDIR], tempGrad)
        for d in xrange(self.velocity.dimension):
            UgradU[XDIR][iCompute] += \
                self.velocity[d][iCompute] * tempGrad[d][iCompute]
        tempGrad = gradOp(self.velocity[YDIR], tempGrad)
        for d in xrange(self.velocity.dimension):
            UgradU[YDIR][iCompute] += \
                self.velocity[d][iCompute] * tempGrad[d][iCompute]
        tempGrad = gradOp(self.velocity[ZDIR], tempGrad)
        for d in xrange(self.velocity.dimension):
            UgradU[ZDIR][iCompute] += \
                self.velocity[d][iCompute] * tempGrad[d][iCompute]

        for d in xrange(self.velocity.dimension):
            result[d][iCompute] += UgradU[d][iCompute]

        ## result = result - nu*\Laplacian u (-g) = gradP/rho
        viscousTerm = npw.zeros_like(self.velocity.data)

        laplacian = diff_op.Laplacian(topo)
        # fd_scheme = FD_C_2((topo.mesh.space_step))
        laplacian.fd_scheme.computeIndices(iCompute)

        for d in xrange(self.velocity.dimension):
            viscousTerm[d] = laplacian(self.velocity[d], viscousTerm[d])
        for d in xrange(self.velocity.dimension):
            viscousTerm[d][iCompute] *= self.viscosity
            result[d][iCompute] -= viscousTerm[d][iCompute]
        ## gravity term : result = result - g with g=(0,-1,-0)
#            if d==1:
#                result[d][iCompute] -= -1.0

        ## baroclinicTerm = -(gradRho/rho) x (gradP/rho)
        gradRho_rho = npw.zeros_like(self.velocity.data)
        gradRho_rho = gradOp(self.density[0], gradRho_rho)

        ## To comment out if the advected scalar is log(rho) :
#        for d in xrange(self.velocity.dimension):
#            gradRho_rho[d][iCompute] = gradRho_rho[d][iCompute] / \
#                                       self.density[0][iCompute]

        baroclinicTerm = npw.zeros_like(result)
        baroclinicTerm[0][iCompute] = - gradRho_rho[1][iCompute] * \
            result[2][iCompute] + gradRho_rho[2][iCompute] \
            * result[1][iCompute]
        baroclinicTerm[1][iCompute] = - gradRho_rho[2][iCompute] * \
            result[0][iCompute] + gradRho_rho[0][iCompute] * \
            result[2][iCompute]
        baroclinicTerm[2][iCompute] = - gradRho_rho[0][iCompute] * \
            result[1][iCompute] + gradRho_rho[1][iCompute] * \
            result[0][iCompute]

        ## vorti(n+1) = vorti(n) + dt * baroclinicTerm
        for d in xrange(self.vorticity.dimension):
            self.vorticity[d][iCompute] += dt * baroclinicTerm[d][iCompute]

        ## velo(n-1) update
        for d in xrange(self.velocity.dimension):
            self.velocity_old[d][iCompute] = self.velocity.data[d][iCompute]
