# -*- coding: utf-8 -*-
"""
@file discrete/differential.py

Discretisation of the differential operators (curl, grad ...)
"""
from hysop.constants import debug
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.numerics.differential_operations import Curl, GradV,\
    DivAdvection
import hysop.tools.numpywrappers as npw
from abc import ABCMeta, abstractmethod
from hysop.numerics.update_ghosts import UpdateGhosts
from hysop.methods_keys import SpaceDiscretisation
try:
    from hysop.f2py import fftw2py
except ImportError:
    from hysop.fakef2py import fftw2py
import hysop.default_methods as default
from hysop.tools.profiler import profile


class Differential(DiscreteOperator):
    __metaclass__ = ABCMeta

    ## @debug
    ## def __new__(cls, *args, **kw):
    ##     return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, invar, outvar, **kwds):
        """
        @param[in] invar : input field
        @param[in,out] outvar : Grad of the input field.
        Warning : must be a field with dim * invar.nbcomponents,
        dim being the domain dimension.
        """
        self.invar = invar
        self.outvar = outvar
        if 'method' not in kwds:
            kwds['method'] = default.DIFFERENTIAL
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(Differential, self).__init__(variables=[invar, outvar],
                                           **kwds)
        self.input = [self.invar]
        self.output = [self.outvar]
        self._synchronize = None
        # connexion to a numerical method
        self._function = None

    @abstractmethod
    def apply(self, simulation=None):
        """
        Abstract interface
        """


class CurlFFT(Differential):
    """
    Compute the curl of a discrete field, using Fourier fftw
    """

    def __init__(self, **kwds):
        super(CurlFFT, self).__init__(**kwds)
        if self.domain.dimension == 3:
            self._apply = self._apply3D
        elif self.domain.dimension == 2:
            raise ValueError('Not yet implemented.')
            #self._apply = self._apply2D

    def apply(self, simulation=None):
        self._apply()

    @debug
    @profile
    def _apply3D(self):
        ghosts_in = self.invar.topology.ghosts()
        ghosts_out = self.outvar.topology.ghosts()
        self.outvar.data[0], self.outvar.data[1], self.outvar.data[2] = \
            fftw2py.solve_curl_3d(self.invar.data[0], self.invar.data[1],
                                  self.invar.data[2], self.outvar.data[0],
                                  self.outvar.data[1], self.outvar.data[2],
                                  ghosts_in, ghosts_out)

    def _apply2D(self):
        ghosts_in = self.invar.topology.ghosts()
        ghosts_out = self.outvar.topology.ghosts()
        self.outvar.data[0], self.outvar.data[1], self.outvar.data[2] = \
            fftw2py.solve_curl_2d(self.invar.data[0], self.invar.data[1],
                                  self.outvar.data[0],
                                  ghosts_in, ghosts_out)

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        fftw2py.clean_fftw_solver(self.outvar.dimension)


class CurlFD(Differential):
    """
    Compute the curl of a discrete field, using finite differences.
    """

    def __init__(self, **kwds):

        super(CurlFD, self).__init__(**kwds)

        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.invar.topology,
                                         self.invar.nb_components)
        self._function = Curl(self.invar.topology, self._rwork,
                              self.method[SpaceDiscretisation])

    def _set_work_arrays(self, rwork, iwork):
        worklength = Curl.getWorkLengths()
        memshape = self.invar.data[0].shape
        if rwork is None:
            self._rwork = [npw.zeros(memshape) for _ in xrange(worklength)]
        else:
            assert len(self._rwork) == worklength
            for wk in self._rwork:
                assert wk.shape == memshape

    @debug
    @profile
    def apply(self, simulation=None):
        self._synchronize(self.invar.data)
        self.outvar.data = self._function(self.invar.data, self.outvar.data)


class GradFD(Differential):
    """
    Compute the grad of a discrete field, using finite differences.
    """

    def __init__(self, **kwds):

        super(GradFD, self).__init__(**kwds)
        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.invar.topology,
                                         self.invar.nb_components)
        dim = self.domain.dimension
        assert self.outvar.nb_components == dim * self.invar.nb_components
        self._function = GradV(self.invar.topology,
                               self.method[SpaceDiscretisation])

    @debug
    @profile
    def apply(self, simulation=None):
        self._synchronize(self.invar.data)
        self.outvar.data = self._function(self.invar.data, self.outvar.data)


class DivAdvectionFD(Differential):
    """
    Compute the grad of a discrete field, using finite differences.
    """

    def __init__(self, **kwds):

        super(DivAdvectionFD, self).__init__(**kwds)
        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.invar.topology,
                                         self.invar.nb_components)
        dim = self.domain.dimension
        assert dim * self.outvar.nb_components == self.invar.nb_components
        self._function = DivAdvection(self.invar.topology,
                                      self.method[SpaceDiscretisation])

    @debug
    @profile
    def apply(self, simulation=None):
        self._synchronize(self.invar.data)
        self.outvar.data = self._function(self.invar.data, self.outvar.data)
