# -*- coding: utf-8 -*-
"""
@file operator/discrete/stretching.py

Discretisation of the stretching operator for different formulations.
"""

from hysop.constants import debug, WITH_GUESS
from hysop.methods_keys import TimeIntegrator, SpaceDiscretisation
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.numerics.integrators.euler import Euler
from hysop.numerics.integrators.runge_kutta2 import RK2
from hysop.numerics.integrators.runge_kutta3 import RK3
from hysop.numerics.integrators.runge_kutta4 import RK4
import hysop.numerics.differential_operations as diff_op
import hysop.tools.numpywrappers as npw
from hysop.numerics.update_ghosts import UpdateGhosts
from hysop.mpi import MPI
from hysop.tools.profiler import profile
from abc import ABCMeta, abstractmethod
import math
ceil = math.ceil


class Stretching(DiscreteOperator):
    """
    Abstract interface to stretching discrete operators.
    Three formulations are available:
    - conservative, see operator.discrete.stretching.Conservative
    - 'Grad(UxW), see operator.discrete.stretching.GradUW
    - Symmetric, see operator.discrete.stretching.Symmetric
    """

    __metaclass__ = ABCMeta

    @debug
    def __init__(self, velocity, vorticity, formulation, **kwds):
        """
        @param velocity : discretization of the velocity field
        @param vorticity : discretization of the vorticity field
        @param formulation : chosen formulation to compute the stretching:
        Conservative or GradUV.
        """
        ## velocity discrete field
        self.velocity = velocity
        ## vorticity discrete field
        self.vorticity = vorticity
        ## Formulation for stretching (divWV or GradVxW)
        self.formulation = formulation

        if 'method' not in kwds:
            import hysop.default_methods as default
            kwds['method'] = default.STRETCHING
        # Work vector used by time-integrator
        self._ti_work = None
        # Work vector used by numerical diff operator.
        self._str_work = None
        super(Stretching, self).__init__(variables=[self.velocity,
                                                    self.vorticity], **kwds)

        self.input = self.variables
        self.output = [self.vorticity]
        # \todo multiresolution case
        assert self.velocity.topology.mesh == self.vorticity.topology.mesh,\
            'Multiresolution case not yet implemented.'

        ## Number of components of the operator (result)
        self.nb_components = 3  # Stretching only in 3D and for vector fields.

        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.velocity.topology,
                                         self.velocity.nb_components
                                         + self.vorticity.nb_components)

        # A function to compute the gradient of a vector field
        # Work vector is provided in input.
        self.strFunc = \
            self.formulation(self.velocity.topology,
                             self._str_work,
                             method=self.method[SpaceDiscretisation])

    def _set_work_arrays(self, rwork=None, iwork=None):

        #v_ind = self.velocity.topology.mesh.iCompute
        shape_v = self.velocity.data[0][...].shape
        ti = self.method[TimeIntegrator]
        # work list length for time-integrator
        work_length_ti = ti.getWorkLengths(3)
        rwork_length = work_length_ti + self.formulation.getWorkLengths()
        # setup for rwork, iwork is useless.
        if rwork is None:
            # ---  Local allocation ---
            self._rwork = []
            for _ in xrange(rwork_length):
                self._rwork.append(npw.zeros(shape_v))
        else:
            assert isinstance(rwork, list), 'rwork must be a list.'
            # --- External rwork ---
            self._rwork = rwork
            assert len(self._rwork) == rwork_length
            for wk in rwork:
                assert wk.shape == shape_v
        self._ti_work = self._rwork[:work_length_ti]
        self._str_work = self._rwork[work_length_ti:]

    def update_ghosts(self):
        """
        Update ghost points values
        """
        self._synchronize(self.velocity.data + self.vorticity.data)

    @abstractmethod
    def apply(self, simulation=None):
        """
        Abstract interface
        """


class Conservative(Stretching):
    """
    Discretisation of the following problem :
    \f{eqnarray*} \frac{\partial\omega}{\partial t} = \nabla.(\omega:v) \f}
    """
    def __init__(self, **kwds):
        super(Conservative, self).__init__(formulation=diff_op.DivWV, **kwds)

        # Time integration scheme.
        def rhs(t, y, result):
            result = self.strFunc(y, self.velocity.data, result)
            return result

        # Create the time integrator
        self.timeIntegrator = \
            self.method[TimeIntegrator](self.nb_components,
                                        self._ti_work,
                                        self.velocity.topology,
                                        f=rhs,
                                        optim=WITH_GUESS)

    @debug
    @profile
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation value for computation."

        # time step
        dt = simulation.timeStep
        # current time
        t = simulation.time
        # Synchronize ghost points of velocity and vorticity
        self._synchronize(self.velocity.data + self.vorticity.data)
        # - Call time integrator -
        # Init workspace with a first evaluation of the
        # rhs of the integrator
        self._ti_work[:self.nb_components] = \
            self.timeIntegrator.f(t, self.vorticity.data,
                                  self._ti_work[:self.nb_components])
        # perform integration and save result in-place
        self.vorticity.data = self.timeIntegrator(t, self.vorticity.data, dt,
                                                  result=self.vorticity.data)


class GradUW(Stretching):
    """
    Discretisation of the following problem:
    \f{eqnarray*} \frac{\partial\omega}{\partial t}=[\nabla(v)][\omega]\f}
    """

    def __init__(self, **kwds):
        super(GradUW, self).__init__(formulation=diff_op.GradVxW, **kwds)

        ## a vector to save diagnostics computed from GradVxW (max div ...)
        self.diagnostics = npw.ones(2)

        # Time integration scheme.
        def rhs(t, y, result):
            result, self.diagnostics =\
                self.strFunc(self.velocity.data, y, result, self.diagnostics)
            return result

        # Create the time integrator
        self.timeIntegrator = \
            self.method[TimeIntegrator](self.nb_components,
                                        self._ti_work,
                                        self.velocity.topology,
                                        f=rhs,
                                        optim=WITH_GUESS)
        ## stability constant
        self.cststretch = 0.
        # Depends on time integration method
        timeint = self.method[TimeIntegrator]
        classtype = timeint.mro()[0]
        if classtype is Euler:
            self.cststretch = 2.0
        elif classtype is RK2:
            self.cststretch = 2.0
        elif classtype is RK3:
            self.cststretch = 2.5127
        elif classtype is RK4:
            self.cststretch = 2.7853

    @debug
    @profile
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation value for computation."

        # time step
        dt = simulation.timeStep
        # current time
        t = simulation.time

        # Synchronize ghost points of velocity
        self._synchronize(self.velocity.data + self.vorticity.data)

        # Compute the number of required subcycles
        ndt, subdt = self.checkStability(dt)
        assert sum(subdt) == dt

        # Init workspace with a first evaluation of the
        # rhs of the integrator
        for i in xrange(ndt):
            self._ti_work[:self.nb_components] = \
                self.timeIntegrator.f(t, self.vorticity.data,
                                      self._ti_work[:self.nb_components])

            # perform integration and save result in-place
            self.vorticity.data = \
                self.timeIntegrator(t, self.vorticity.data, subdt[i],
                                    result=self.vorticity.data)

    def checkStability(self, dt):
        """
        Computes a stability condition depending on some
        diagnostics (from GradVxW)
        @param current time step
        @return int,float : the number of required subcycles and
        the subcycle time-step.
        """
        dt_stab = min(dt, self.cststretch / self.diagnostics[1])
        nb_cycles = int(ceil(dt / dt_stab))
        subdt = npw.zeros((nb_cycles))
        subdt[:] = dt_stab
        subdt[-1] = dt - (nb_cycles - 1) * dt_stab
        return nb_cycles, subdt


class Symmetric(Stretching):
    """
    Discretisation of the following problem :
    \f{eqnarray*} \frac{\partial\omega}{\partial t} = \frac{1}{2}(\nabla(v) + \nabla(v)^T)\omega \f}
    """
    def __init__(self, **kwds):
        super(Symmetric, self).__init__(formulation=diff_op.DivWV, **kwds)

        # Time integration scheme.
        def rhs(t, y, result):
            result = self.strFunc(y, self.velocity.data, result)
            return result

        # Create the time integrator
        self.timeIntegrator = \
            self.method[TimeIntegrator](self.nb_components,
                                        self._ti_work,
                                        self.velocity.topology,
                                        f=rhs,
                                        optim=WITH_GUESS)

    @debug
    @profile
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation value for computation."

        # time step
        dt = simulation.timeStep
        # current time
        t = simulation.time
        # Synchronize ghost points of velocity and vorticity
        self._synchronize(self.velocity.data + self.vorticity.data)
        # - Call time integrator -
        # Init workspace with a first evaluation of the
        # rhs of the integrator
        self._ti_work[:self.nb_components] = \
            self.timeIntegrator.f(t, self.vorticity.data,
                                  self._ti_work[:self.nb_components])
        # perform integration and save result in-place
        self.vorticity.data = self.timeIntegrator(t, self.vorticity.data, dt,
                                                  result=self.vorticity.data)


