# -*- coding: utf-8 -*-
"""
@file operator/stretching.py

Computation of stretch. term in Navier-Stokes.

"""
from hysop.constants import debug
from hysop.methods_keys import TimeIntegrator, Formulation, \
    SpaceDiscretisation
from hysop.numerics.finite_differences import FD_C_4
from hysop.operator.computational import Computational
from hysop.operator.continuous import opsetup
from hysop.operator.discrete.stretching import Conservative, GradUW
from abc import ABCMeta


class Stretching(Computational):
    """
    \todo write latex formulas
    """
    __metaclass__ = ABCMeta

    @debug
    def __init__(self, velocity, vorticity, **kwds):
        """
        Create a Stretching operator from given
        velocity and vorticity variables.
        @param velocity field
        @param vorticity field
        Default solving method : finite differences, 4th order, in space
        and Runge-Kutta 3 in time.
        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(Stretching, self).__init__(variables=[velocity, vorticity],
                                         **kwds)
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        ## Numerical methods for time and space discretization

        if self.method is None:
            import hysop.default_methods as default
            self.method = default.STRETCHING
        assert Formulation in self.method.keys()
        assert SpaceDiscretisation in self.method.keys()
        assert TimeIntegrator in self.method.keys()

        ## Formulation used for the stretching equation.
        ## Default = conservative form.
        if self.method[Formulation] == "GradUW":
            self.formulation = GradUW
        else:
            self.formulation = Conservative

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]

    def get_work_properties(self):
        if not self._is_discretized:
            msg = 'The operator must be discretized '
            msg += 'before any call to this function.'
            raise RuntimeError(msg)
        #v_ind = self.variables[self.velocity].mesh.iCompute
        vd = self.discreteFields[self.velocity]
        shape_v = vd[0][...].shape
        ti = self.method[TimeIntegrator]
        rwork_length = ti.getWorkLengths(3)
        import hysop.numerics.differential_operations as diff_op
        if self.formulation is GradUW:
            rwork_length += diff_op.GradVxW.getWorkLengths()
        elif self.formulation is Conservative:
            rwork_length += diff_op.DivWV.getWorkLengths()
        res = {'rwork': [], 'iwork': None}
        for _ in xrange(rwork_length):
            res['rwork'].append(shape_v)
        return res

    def discretize(self):
        if self.method[SpaceDiscretisation] is FD_C_4:
            nbGhosts = 2
        else:
            raise ValueError("Unknown method for space discretization of the\
                stretching operator.")

        super(Stretching, self)._standard_discretize(nbGhosts)

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        self.discrete_op =\
            self.formulation(velocity=self.discreteFields[self.velocity],
                             vorticity=self.discreteFields[self.vorticity],
                             method=self.method, rwork=rwork, iwork=iwork)
        self._is_uptodate = True
