"""
@file hysop.operator.tests.test_differential
Tests for differential operators.
"""
import numpy as np
from hysop.domain.box import Box
from hysop.fields.continuous import Field
import hysop.tools.numpywrappers as npw
#from hysop.numerics.differential_operations import DivT, GradVxW


# Domain and topologies definitions
nb = 65
import math
Lx = Ly = Lz = 2. * math.pi
box = Box(length=[Lx, Ly, Lz], origin=[0., 0., 0.])
box2d = Box(length=[Lx, Ly], origin=[0., 0.])
from hysop.tools.parameters import Discretization
d3D = Discretization([nb, nb, nb], [2, 2, 2])
d2D = Discretization([nb, nb], [2, 2])
d3D_nog = Discretization([nb, nb, nb])

box2 = Box(length=[2. * Lx, Ly, Lz], origin=[0., 0., 0.])
d3D_2 = Discretization([129, nb, nb])
cos = np.cos
sin = np.sin

def velocity_f(res, x, y, z, t):
    res[0][...] = sin(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


def vorticity_f(res, x, y, z, t):
    res[0][...] = - cos(x) * sin(y) * sin(z)
    res[1][...] = - sin(x) * cos(y) * sin(z)
    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
    return res


def velocity_f2d(res, x, y, t):
    res[0][...] = sin(x) * cos(y)
    res[1][...] = - cos(x) * sin(y)
    return res


def vorticity_f2d(res, x, y, t):
    res[0][...] = 2. * sin(x) * sin(y)
    return res


def grad_velo(res, x, y, z, t):
    res[0][...] = cos(x) * cos(y) * cos(z)
    res[1][...] = -sin(x) * sin(y) * cos(z)
    res[2][...] = -sin(x) * cos(y) * sin(z)
    res[3][...] = sin(x) * sin(y) * cos(z)
    res[4][...] = - cos(x) * cos(y) * cos(z)
    res[5][...] = cos(x) * sin(y) * sin(z)
    res[6][...] = 0.0
    res[7][...] = 0.0
    res[8][...] = 0.0
    return res


def callOp(DiffOperator, ref_formula, discretization,
           op_dim=box.dimension, method=None,
           order=4, dom=box, vform=velocity_f):
    """Basic test for Grad operator using FD scheme
    (comparison with periodic analytical solution)
    """
    # Velocity and result fields
    velo = Field(domain=dom, formula=vform, is_vector=True)
    result = Field(domain=dom, nb_components=op_dim)
    # Curl operator
    Op = DiffOperator(velo, result, discretization=discretization,
                      method=method)

    Op.discretize()
    topo = Op.discreteFields[velo].topology

    # Reference field
    ref = Field(domain=dom, formula=ref_formula, nb_components=op_dim)
    ref_d = ref.discretize(topo)
    ref.initialize(topo=topo)
    res_d = result.discreteFields[topo]

    # initialize velocity ...
    velo.initialize(topo=topo)
    Op.setup()

    # Compute curl
    Op.apply()

    # Compare results with reference
    ind = topo.mesh.iCompute
    err = npw.zeros((dom.dimension))
    for i in xrange(dom.dimension):
        err[i] = (dom.length[i] / (discretization.resolution[i] - 1)) ** order
    print ('==============')
    print (str(DiffOperator) + ' test')
    print ('==============')
    for i in xrange(result.nb_components):
        print ('err = O(h**order) =', err[i % dom.dimension])
        print (np.max(np.abs(res_d[i][ind] - ref_d[i][ind])))
        assert np.allclose(res_d[i][ind], ref_d[i][ind],
                           rtol=err[i % dom.dimension])
    Op.finalize()


def test_CurlFD():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_4
    from hysop.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Curl, vorticity_f, method=method, discretization=d3D)


def test_CurlFD2():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_2
    from hysop.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_2}
    callOp(Curl, vorticity_f, method=method, order=2, discretization=d3D)


def test_CurlFD2D():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_4
    from hysop.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Curl, vorticity_f2d, method=method, discretization=d2D,
           dom=box2d, op_dim=1, vform=velocity_f2d)


def test_CurlFFT():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.operator.differential import Curl
    method = {SpaceDiscretisation: 'fftw'}
    callOp(Curl, vorticity_f, method=method, order=6, discretization=d3D_nog)


#def test_CurlFFT_ghosts():
#    from hysop.methods_keys import SpaceDiscretisation
#    from hysop.operator.differential import Curl
#    method = {SpaceDiscretisation: 'fftw'}
#    callOp(Curl, vorticity_f, method=method, order=6, discretization=d3D)


#def test_CurlFFT_2():
#    from hysop.methods_keys import SpaceDiscretisation
#    from hysop.operator.differential import Curl
#    method = {SpaceDiscretisation: 'fftw'}
#    discr = Discretization([129, nb, nb])
#    callOp(Curl, vorticity_f, method=method, order=6, dom=box2,
#           discretization=discr)


def test_Grad():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_2
    method = {SpaceDiscretisation: FD_C_2}
    from hysop.operator.differential import Grad
    callOp(Grad, grad_velo, op_dim=9, method=method, order=2,
           discretization=d3D)


def test_Grad2():
    from hysop.operator.differential import Grad
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_4
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Grad, grad_velo, op_dim=9, method=method, discretization=d3D)


def test_CurlFD_2():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_4
    from hysop.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Curl, vorticity_f, method=method, dom=box2, discretization=d3D_2)


def test_CurlFD2_2():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_2
    from hysop.operator.differential import Curl
    method = {SpaceDiscretisation: FD_C_2}
    callOp(Curl, vorticity_f, method=method, order=2, dom=box2,
           discretization=d3D_2)


def test_Grad_2():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_2
    method = {SpaceDiscretisation: FD_C_2}
    from hysop.operator.differential import Grad
    callOp(Grad, grad_velo, op_dim=9, method=method, order=2, dom=box2,
           discretization=d3D_2)


def test_Grad2_2():
    from hysop.operator.differential import Grad
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_4
    method = {SpaceDiscretisation: FD_C_4}
    callOp(Grad, grad_velo, op_dim=9, method=method, dom=box2,
           discretization=d3D_2)


def test_CurlFD_work():
    from hysop.methods_keys import SpaceDiscretisation
    from hysop.methods import FD_C_4
    from hysop.operator.differential import Curl
    # Velocity and result fields
    velo = Field(domain=box, formula=velocity_f, is_vector=True)
    result = Field(domain=box, nb_components=box.dimension)
    # Curl operator
    Op = Curl(velo, result, discretization=d3D)

    Op.discretize()
    topo = Op.discreteFields[velo].topology

    # Reference field
    ref = Field(domain=box, formula=vorticity_f, nb_components=box.dimension)
    ref_d = ref.discretize(topo)
    ref.initialize(topo=topo)
    res_d = result.discreteFields[topo]

    # initialize velocity ...
    velo.initialize(topo=topo)
    wk_p = Op.get_work_properties()
    rwork = []
    wk_length = len(wk_p['rwork'])
    for i in xrange(wk_length):
        memshape = wk_p['rwork'][i]
        rwork.append(npw.zeros(memshape))

    Op.setup(rwork=rwork)

    # Compute curl
    Op.apply()

    # Compare results with reference
    ind = topo.mesh.iCompute
    err = npw.zeros((box.dimension))
    for i in xrange(box.dimension):
        err[i] = (box.length[i] / (d3D.resolution[i] - 1)) ** 4
    for i in xrange(result.nb_components):
        print ('err = O(h**order) =', err[i % box.dimension])
        assert np.allclose(res_d[i][ind], ref_d[i][ind],
                           rtol=err[i % box.dimension])
    Op.finalize()


# This may be useful to run mpi tests
if __name__ == "__main__":
    test_CurlFD()
    test_CurlFD2()
    test_CurlFD2D()
    test_CurlFFT()
    test_Grad()
    test_Grad2()
    test_CurlFD_2()
    test_CurlFD2_2()
    #test_CurlFFT_2()
    test_Grad_2()
    test_Grad2_2()
    test_CurlFD_work()
