# -*- coding: utf-8 -*-
from hysop.domain.subsets import Sphere
from hysop.operator.penalization import PenalizeVorticity
from hysop.problem.simulation import Simulation
from hysop.tools.parameters import Discretization, IOParams
from hysop.mpi.topology import Cartesian
import numpy as np
import os
from hysop import Field, Box
from hysop.operator.hdf_io import HDF_Reader
from hysop.operator.drag_and_lift import MomentumForces, NocaForces


def v2d(res, x, y, t):
    res[0][...] = 1.
    res[1][...] = 1.
    return res


def s2d(res, x, y, t):
    res[0][...] = 1.
    return res


def v3d(res, x, y, z, t):
    res[0][...] = 1.
    res[1][...] = 1.
    res[2][...] = 1.
    return res


def w3d(res, x, y, z, t):
    res[0][...] = 1.
    res[1][...] = 1.
    res[2][...] = 1.
    return res


def s3d(res, x, y, z, t):
    res[0][...] = 1.
    return res


Nx = 128
Ny = 96
Nz = 102
g = 2


ldef = [0.3, 0.4, 1.0]
discr3D = Discretization([Nx + 1, Ny + 1, Nz + 1], [g, g, g])
discr2D = Discretization([Nx + 1, Ny + 1], [g, g])
xdom = np.asarray([0.1, -0.3, 0.5])
import math
ldom = np.asarray([math.pi * 2., ] * 3)
xdef = xdom + 0.2
xpos = ldom * 0.5
xpos[-1] += 0.1
from hysop.mpi import main_size
working_dir = os.getcwd() + '/'


def init_vort(discr, fileref):
    Cartesian.reset_counter()
    dim = len(discr.resolution)
    dom = Box(dimension=dim, length=ldom[:dim],
              origin=xdom[:dim])
    topo = dom.create_topology(discr)
    wref = Field(domain=topo.domain, name='vortiref', is_vector=dim == 3)
    #    scalRef.hdf_load(topo, iop, restart=0)
    veloref = Field(domain=topo.domain, name='veloref', is_vector=True)
    # Read a reference file
    iop = IOParams(working_dir + fileref)
    reader = HDF_Reader(variables={wref: topo, veloref: topo},
                        io_params=iop, restart=0)
    reader.discretize()
    reader.setup()
    #reader.apply()
    #reader.finalize()
    wdref = wref.discretize(topo)
    vdref = veloref.discretize(topo)
    return topo, wdref, vdref


def check_drag(penal, wref, vref, vorti, velo):
    penal.discretize()
    penal.setup()
    topo = penal.variables[vorti]
    vorti.initialize(topo=topo)
    velo.initialize(topo=topo)
    vd = velo.discretize(topo)
    wd = vorti.discretize(topo)
    simu = Simulation(nbIter=3)
    penal.apply(simu)
    ind = topo.mesh.iCompute
    for d in xrange(vd.nb_components):
        assert np.allclose(vd.data[d][ind], vref.data[d][ind])
    for d in xrange(wd.nb_components):
        assert np.allclose(wd.data[d][ind], wref.data[d][ind])


def test_momentum_forces_3d():
    """
    Compute drag/lift in 3D, flow around a sphere
    """
    topo, wref, vref = init_vort(discr3D, 'drag_3d_sphere')
    # Velocity field
    velo = Field(domain=topo.domain, formula=v3d, name='Velo', is_vector=True)
    # obstacle
    rd = ldom[0] * 0.2
    hsphere = Sphere(parent=topo.domain, origin=xpos, radius=rd)
    hsphere.discretize(topo)
    # Forces
    dg = MomentumForces(velocity=velo, discretization=topo,
                        obstacles=[hsphere], penalisation_coeff=[1e8])
    dg.discretize()
    dg.setup()
    simu = Simulation(nbIter=3)
    dg.apply(simu)


def build_noca(formulation):
    """
    Compute drag/lift in 3D, flow around a sphere
    """
    topo, wref, vref = init_vort(discr3D, 'drag_3d_sphere')
    dim = topo.domain.dimension
    # Velocity field
    velo = Field(domain=topo.domain, formula=v3d, name='Velo', is_vector=True)
    vorti = Field(domain=topo.domain, formula=w3d,
                  name='Vorti', is_vector=dim == 3)
    # obstacle
    rd = ldom[0] * 0.2
    hsphere = Sphere(parent=topo.domain, origin=xpos, radius=rd)
    # penalisation
    penal = PenalizeVorticity(velocity=velo, vorticity=vorti,
                              discretization=topo,
                              obstacles=[hsphere], coeff=1e8)
    penal.discretize()
    penal.setup()
    # Forces
    dg = NocaForces(velocity=velo, vorticity=vorti, discretization=topo,
                    nu=0.3, obstacles=[hsphere], formulation=formulation)
    dg.discretize()
    dg.setup()
    simu = Simulation(nbIter=3)
    penal.apply(simu)
    dg.apply(simu)


def test_noca1():
    build_noca(1)


def test_noca2():
    build_noca(2)


def test_noca3():
    build_noca(3)

# This may be useful to run mpi tests
if __name__ == "__main__":
    test_momentum_forces_3d()
    #test_noca1()
    #test_noca2()
    #test_noca3()
