from hysop.mpi import MPI
import hysop.tools.numpywrappers as npw
import inspect
import numpy as np
import re
ftime = MPI.Wtime


def profile(f):
    """
    Decorator for function profiling.
    The function must be an instance method.
    """
    def deco(*args, **kwargs):
        """args[0] contains the object"""
        t0 = ftime()
        res = f(*args, **kwargs)
        args[0].profiler[f.func_name] += ftime() - t0
        return res
    return deco


class FProfiler(object):
    """
    Class for time measurments on the fly.
    The objects can be linked to a class method.
    """
    def __init__(self, fname):
        ## Function name
        self.fname = fname
        ## Total exectution time
        self.t = 0.
        ## Call numbers
        self.n = 0

    def get_name(self):
        return self.fname

    def __iadd__(self, t):
        self.t += t
        self.n += 1
        return self

    def __str__(self):
        if self.n > 0:
            s = "- {0} : {1} s ({2} calls)".format(
                self.fname, self.t, self.n)
        else:
             s = ""
        return s


class Profiler(object):
    def __init__(self, obj, comm):
        self.summary = {}
        self.table = []
        self._comm = comm
        self._comm_size = comm.Get_size()
        self._comm_rank = comm.Get_rank()
        for _ in xrange(self._comm_size):
            self.table.append([])
        self._elems = {}
        self._obj = obj
        self._l = 1
        self.all_times = None

    def down(self, l):
        self._l = l + 1

    def get_name(self):
        try:
            _name = self._obj.name
        except AttributeError:
            if isinstance(self._obj, str):
                _name = self._obj
            else:
                _name = 'unknown'
        return _name

    def __iadd__(self, other):
        self._elems[other.get_name()] = other
        return self

    def __setitem__(self, key, value):
        self._elems[key] = value

    def __getitem__(self, item):
        try:
            return self._elems[item]
        except KeyError:
            self._elems[item] = FProfiler(item)
            return self._elems[item]

    def __str__(self):
        from hysop.mpi.main_var import main_rank
        if len(self.summary) > 0:
            s = ""
            if self._l == 1:
                s += "\n[{0}]".format(main_rank)
            s += '- ' + self.get_name() + ":"
            for k in sorted(self.summary.keys()):
                if len(str(self.summary[k])) > 0:
                    s += "\n[{0}]".format(main_rank)
                    s += '  ' * self._l + str(self.summary[k])
        else:
            s = ""
        return s

    def write(self, prefix='', hprefix='', with_head=True):
        if prefix != '' and prefix[-1] != ' ':
            prefix += ' '
        if hprefix != '' and hprefix[-1] != ' ':
            hprefix += ' '
        if self._comm_rank == 0:
            s = ""
            h = hprefix + "Rank"
            for r in xrange(self._comm_size):
                s += prefix + "{0}".format(r)
                for i in xrange(len(self.all_names)):
                    s += " {0}".format(self.all_times[i][r])
                s += "\n"
            s += prefix + "-1"
            for i in xrange(len(self.all_names)):
                h += ' ' + self.all_names[i]
                s += " {0}".format(self.all_times[i][self._comm_size])
            h += "\n"
            if with_head:
                s = h + s
            print s

    def summarize(self):
        from hysop.fields.continuous import Field
        self.summary = {}
        try:
            self._obj.get_profiling_info()
        except AttributeError:
            pass
        i = 0
        for k in self._elems.keys():
            try:
                self.summary[self._elems[k].t] = self._elems[k]
            except AttributeError:
                i += 1
                self._elems[k].down(self._l)
                self._elems[k].summarize()
                if isinstance(self._elems[k]._obj, Field):
                    self.summary[1e10 * i] = self._elems[k]
                else:
                    self.summary[1e8 * i] = self._elems[k]

        for k in sorted(self._elems.keys()):
            if isinstance(self._elems[k], FProfiler):
                self.table[self._comm_rank].append(
                    (self.get_name() + '.' + k,
                     self._elems[k].t, self._elems[k].n))
        for k in sorted(self._elems.keys()):
            if isinstance(self._elems[k], Profiler) and \
               not isinstance(self._elems[k]._obj, Field):
                for e in self._elems[k].table[self._comm_rank]:
                    self.table[self._comm_rank].append(
                        (self.get_name() + '.' + e[0], e[1], e[2]))
            if isinstance(self._elems[k], Profiler) and \
               isinstance(self._elems[k]._obj, Field):
                for e in self._elems[k].table[self._comm_rank]:
                    self.table[self._comm_rank].append(
                        (self.get_name() + '.' + e[0], e[1], e[2]))

        if self._l == 1 and self.all_times is None:
            nb = len(self.table[self._comm_rank])
            self.all_times = npw.zeros((nb, self._comm_size + 1))
            self.all_call_nb = npw.int_zeros((nb, self._comm_size + 1))
            self.all_names = [None] * nb
            for i in xrange(nb):
                self.all_names[i] = self.table[self._comm_rank][i][0]
                self.all_times[i, self._comm_rank] = self.table[self._comm_rank][i][1]
                self.all_call_nb[i, self._comm_rank] = self.table[self._comm_rank][i][2]
            for p in xrange(self._comm_size):
                self._comm.Bcast(self.all_times[:, p], root=p)
                self._comm.Bcast(self.all_call_nb[:, p], root=p)
            self.all_times[:, self._comm_size] = np.sum(
                self.all_times[:, :self._comm_size], axis=1)
            self.all_call_nb[:, self._comm_size] = np.sum(
                self.all_call_nb[:, :self._comm_size], axis=1)
